@@ -2,17 +2,19 @@ __author__ = 'Altertech, https://www.altertech.com/'
 __copyright__ = 'Copyright (C) 2019 Altertech'
 __license__ = 'MIT'
 
-__version__ = '0.2.21'
+__version__ = '0.3.3'
 
 from sqlalchemy.exc import IntegrityError
 from cachetools import TTLCache
 from itertools import groupby
 from .currencies import currencies
 
-rate_cache = TTLCache(maxsize=100, ttl=2)
+rate_cache = TTLCache(maxsize=100, ttl=5)
 
 asset_precision_cache = {}
 
+restrict_assets_to_currencies = False
+
 # financial assets
 ACCOUNT_CREDIT = 0
 ACCOUNT_CASH = 1
@@ -357,14 +359,14 @@ def init(db=None, **kwargs):
             "EUR/USD" pair exists but no USD/EUR, use 1 / "EUR/USD"
         rate_allow_cross: if exchange rate is not found, allow finac to look
             for the nearest cross-asset rate
-        rate_ttl: set rate cache ttl (default: 2 sec)
+        rate_ttl: set rate cache ttl (default: 5 sec)
         full_transaction_update: allow updating transaction date and amount
         base_asset: default base asset. Default is "USD"
         date_format: default date format in statements
     """
     for k, v in kwargs.items():
         if k == 'rate_ttl':
-            rate_cache.__setstate__({'_TTLCache__ttl': 600})
+            rate_cache.__setstate__({'_TTLCache__ttl': v})
         else:
             if not hasattr(config, k):
                 raise RuntimeError('Parameter {} is invalid'.format(k))
@@ -374,10 +376,9 @@ def init(db=None, **kwargs):
         db_uri = db
         if db_uri.find('.db') != -1:
             db_uri = 'sqlite:///' + os.path.expanduser(db_uri)
-        elif not db_uri.__contains__('mysql+pymysql'):
-            raise RuntimeError(
-                'DB URI configuration: mysql+pymysql://user:pass@host/db_name')
         _db.engine = get_db_engine(db_uri)
+        _db.use_lastrowid = db_uri.startswith('sqlite') or db_uri.startswith(
+            'mysql')
         init_db(_db.engine)
 
 
@@ -392,9 +393,10 @@ def asset_create(asset, precision=2):
             operations. Default is 2 digits
     """
     asset = asset.upper()
-    if asset not in [c['cc'] for c in currencies]:
-        logger.error('Currency {} not found'.format(asset))
-        raise ResourceNotFound
+    if restrict_assets_to_currencies:
+        if asset not in [c['cc'] for c in currencies]:
+            logger.error('Currency {} not found'.format(asset))
+            raise ResourceNotFound
     logger.info('Creating asset {}'.format(asset))
     try:
         get_db().execute(sql("""
@@ -427,7 +429,7 @@ def asset_list_rates(asset=None, start=None, end=None):
     """
     List asset rates
 
-    Currency can be specified either as code, or as pair "code/code"
+    Asset can be specified either as code, or as pair "code/code"
 
     If asset is not specified, "end" is used as date to get rates for all
     assets
@@ -443,12 +445,12 @@ def asset_list_rates(asset=None, start=None, end=None):
         if asset.find('/') != -1:
             asset_from, asset_to = _safe_format(asset.split('/'))
             cond += (' and ' if cond else
-                     '') + 'cf.code = "{}" and ct.code = "{}"'.format(
+                     '') + 'cf.code = \'{}\' and ct.code = \'{}\''.format(
                          asset_from, asset_to)
         else:
-            cond += (' and ' if cond else
-                     '') + '(cf.code = "{code}" or ct.code = "{code}")'.format(
-                         code=asset)
+            cond += (' and ' if cond else ''
+                    ) + '(cf.code = \'{code}\' or ct.code = \'{code}\')'.format(
+                        code=asset)
         r = get_db().execute(
             sql("""
             select cf.code as asset_from,
@@ -472,8 +474,6 @@ def asset_list_rates(asset=None, start=None, end=None):
                     as1.asset_to_id as t,
                     max(as1.d) as m
                 from asset_rate as as1
-                    inner join asset_rate as as2
-                    on as1.asset_from_id=as2.asset_from_id
                 where as1.d<=:d group by fr, t)
             as s1
                 join asset as a1 on a1.id=fr
@@ -505,7 +505,7 @@ def asset_delete(asset):
     if not get_db().execute(sql("""
     delete from asset where code=:code"""),
                             code=asset.upper()).rowcount:
-        logger.error('Currency {} not found'.format(asset.upper()))
+        logger.error('Asset {} not found'.format(asset.upper()))
         raise ResourceNotFound
 
 
@@ -527,7 +527,9 @@ def asset_set_rate(asset_from, asset_to=None, value=None, date=None):
         value = asset_to
         asset_to = None
     if value is None:
-        raise ValueError('Currency rate value is not specified')
+        raise ValueError('Asset rate value is not specified')
+    else:
+        value = parse_number(value)
     if date is None:
         date = int(time.time())
     else:
@@ -574,7 +576,7 @@ def asset_delete_rate(asset_from, asset_to=None, date=None):
                             f=asset_from.upper(),
                             t=asset_to.upper(),
                             d=date).rowcount:
-        logger.error('Currency rate {}/{} for {} not found'.format(
+        logger.error('Asset rate {}/{} for {} not found'.format(
             asset_from.upper(), asset_to.upper(), format_date(date)))
         raise ResourceNotFound
 
@@ -709,39 +711,45 @@ def account_create(account,
     else:
         passive = val_to_boolean(passive)
     db = get_db()
+    account = account.upper()
+    asset = asset.upper()
+    if not db.execute(sql("""select id from asset where code=:code"""),
+                      code=asset).fetchone():
+        raise ResourceNotFound('asset {}'.format(asset))
     dbt = db.begin()
-    logger.info('Creating account {}, asset: {}'.format(account.upper(),
-                                                        asset.upper()))
+    logger.info('Creating account {}, asset: {}'.format(account, asset))
     try:
         r = db.execute(sql("""
         insert into account(code, note, tp, passive, asset_id, max_overdraft,
         max_balance) values
         (:code, :note, :tp, :passive,
             (select id from asset where code=:asset),
-            :max_overdraft, :max_balance)"""),
-                       code=account.upper(),
+            :max_overdraft, :max_balance) {}""".format(
+            '' if _db.use_lastrowid else 'returning id')),
+                       code=account,
                        note=note,
                        tp=tp_id,
                        passive=passive,
-                       asset=asset.upper(),
+                       asset=asset,
                        max_overdraft=max_overdraft,
                        max_balance=max_balance)
+        acc_id = r.lastrowid if _db.use_lastrowid else r.fetchone().id
         db.execute(sql("""
             insert into transact(account_debit_id, amount, d_created, d) values
             (:account_id, 0, 0, 0)
             """),
-                   account_id=r.lastrowid)
+                   account_id=acc_id)
         db.execute(sql("""
             insert into transact(account_credit_id, amount, d_created, d) values
             (:account_id, 0, 0, 0)
             """),
-                   account_id=r.lastrowid)
+                   account_id=acc_id)
         dbt.commit()
     except IntegrityError:
         dbt.rollback()
-        raise ResourceAlreadyExists(account.upper())
+        raise ResourceAlreadyExists(account)
     except:
-        logger.error('Unable to create account {}'.format(account.upper()))
+        logger.error('Unable to create account {}'.format(account))
         dbt.rollback()
         raise
 
@@ -1141,22 +1149,24 @@ def _transaction_move(dt=None,
             completion_date = date
     else:
         completion_date = parse_date(completion_date)
-    return db.execute(sql("""
+    r = db.execute(sql("""
     insert into transact(account_credit_id, account_debit_id, amount, tag,
     note, d_created, d, chain_transact_id) values
     (
     (select id from account where code=:ct),
     (select id from account where code=:dt),
     :amount, :tag, :note, :d_created, :d, :chain_id)
-    """),
-                      ct=ct,
-                      dt=dt,
-                      amount=amount,
-                      tag=tag,
-                      note=note,
-                      d_created=date,
-                      d=completion_date,
-                      chain_id=chain_transact_id).lastrowid
+    {}
+    """.format('' if _db.use_lastrowid else 'returning id')),
+                   ct=ct,
+                   dt=dt,
+                   amount=amount,
+                   tag=tag,
+                   note=note,
+                   d_created=date,
+                   d=completion_date,
+                   chain_id=chain_transact_id)
+    return r.lastrowid if _db.use_lastrowid else r.fetchone().id
 
 
 @core_method
@@ -1237,7 +1247,7 @@ def transaction_move(dt=None,
                                       date=date)
 
             else:
-                raise ValueError('Currency mismatch')
+                raise ValueError('Asset mismatch')
             tid1 = _transaction_move(
                 ct=ct,
                 amount=format_amount(amount /
@@ -1464,7 +1474,7 @@ def account_statement(account, start=None, end=None, tag=None, pending=True):
         tag = _safe_format(tag) if isinstance(tag,
                                               (list,
                                                tuple)) else [_safe_format(tag)]
-        tf = ['tag = "{}"'.format(t) for t in tag]
+        tf = ['tag = \'{}\''.format(t) for t in tag]
         tags = ' or '.join(tf)
         cond += (' and ' if cond else '') + '({tags})'.format(tags=tags)
     r = get_db().execute(sql("""
@@ -1598,14 +1608,14 @@ def account_list(asset=None,
                 first = False
             else:
                 cond += ' or '
-            cond += 'asset.code = "{}"'.format(_safe_format(a.upper()))
+            cond += 'asset.code = \'{}\''.format(_safe_format(a.upper()))
         cond += ')'
     else:
         cond += (' and ' if cond else '') + 'account.tp <= 1000'
     dts = parse_date(date) if date else parse_date()
-    cond += (' and ' if cond else '') + 'transact.d_created <= "{}"'.format(dts)
+    cond += (' and ' if cond else '') + 'transact.d_created <= {}'.format(dts)
     if code:
-        cond += (' and ' if cond else '') + 'account.code like "{}"'.format(
+        cond += (' and ' if cond else '') + 'account.code like \'{}\''.format(
             _safe_format(code))
     passive = val_to_boolean(passive)
     if passive is True:
@@ -1634,7 +1644,8 @@ def account_list(asset=None,
                     left join account on account.id=transact.account_debit_id
                     join asset on asset.id=account.asset_id
                     where d is not null and {cond_d}
-                        group by account.code
+                        group by account.code, account.note,
+                            account.passive, asset.code, account.tp
                 union
                 select -1*sum(amount) as balance,
                     account.code as account,
@@ -1646,7 +1657,8 @@ def account_list(asset=None,
                     left join account on account.id=transact.account_credit_id
                     join asset on asset.id=account.asset_id
                     where {cond}
-                            group by account.code
+                            group by account.code, account.note,
+                            account.passive, asset.code, account.tp
                 ) as templist
                     group by account, note, passive, templist.asset, templist.tp
             {oby}
@@ -1656,7 +1668,8 @@ def account_list(asset=None,
     while True:
         d = r.fetchone()
         if not d: break
-        if hide_empty is False or d.balance:
+        # if zero is not a "real zero" - consider x < 0.000001 is zero
+        if hide_empty is False or (abs(d.balance) > 0.000001):
             row = OrderedDict()
             for i in ('account', 'type', 'passive', 'note', 'asset', 'balance'):
                 if i == 'type':
@@ -1835,14 +1848,14 @@ def _account_summary(balance_type,
         'transact.d is not null and ' if _safe_format(balance_type) ==
         'debit' else '')
     if account:
-        cond += (' and ' if cond else '') + 'account.code = "{}"'.format(
+        cond += (' and ' if cond else '') + 'account.code = \'{}\''.format(
             _safe_format(account))
     if asset:
-        cond += (' and ' if cond else '') + 'asset.code = "{}"'.format(
+        cond += (' and ' if cond else '') + 'asset.code = \'{}\''.format(
             _safe_format(asset))
     if date:
         dts = parse_date(_safe_format(date))
-        cond += (' and ' if cond else '') + 'transact.d <= "{}"'.format(dts)
+        cond += (' and ' if cond else '') + 'transact.d <= {}'.format(dts)
     if tp:
         if isinstance(tp, int):
             tp_id = tp
@@ -1898,7 +1911,7 @@ def account_balance(account=None, tp=None, base=None, date=None,
         tp = [k for k in ACCOUNT_TYPE_IDS if ACCOUNT_TYPE_IDS[k] <= 1000]
     cond = "transact.deleted is null"
     dts = parse_date(_safe_format(date)) if date else parse_date()
-    cond += (' and ' if cond else '') + 'transact.d_created <= "{}"'.format(dts)
+    cond += (' and ' if cond else '') + 'transact.d_created <= {}'.format(dts)
     balance = None
     if account:
         acc_info = account_info(account)
